/*
sjis Version 1.04
Trying to convert raw keys into Shift_JIS keys
Copyright 2015 Sogaya
Please enter "sjis -V" for usage.

Watcom

*/

#define VERSION 1.04
#define BINARY_FILE "sjis.bin"
#define BUFFER_SIZE 256
#define CHARACTER_COUNT 8
#define FROM_TO " -> "
#define FULL 0x100
#define KANA_KUTEN 0xA1
#define SALT_ARRAY_SIZE 82
#define TRUE 1
#define VOYA_LOG "voya.log"
#include <stdio.h>
#include <string.h>

typedef int BOOL;
int byte_table[FULL][FULL];

int charToInteger(char c) {
    unsigned char uc = (unsigned char)c;
    int i = (int)uc;
    return i;
}

void chopped(char *s) {
    char *p;

/* Removing the new-line character */
    p = strrchr(s, '\n');
    if(p != NULL) {
        *p = '\0';
    }
}

BOOL zenkaku_fragment(char *buffer) {
/* Checking if the line ends with a separated byte of kanji. */
    BOOL zenkaku = 0; /* FALSE */;
    char line[BUFFER_SIZE];
    int c, k;

    strcpy(line, buffer);
    for(k = 0; line[k] != '\0'; k++) {
        c = charToInteger(line[k]);

        if(zenkaku) {
            zenkaku = 0; /* FALSE */;
            continue;
        }

        if(0x81 <= c && c <= 0xFC) {
            if(KANA_KUTEN <= c && c <= 0xDF) { /*  */
                zenkaku = 0; /* FALSE */;
            } else {
                zenkaku = TRUE;
            }

        } else {
            zenkaku = 0; /* FALSE */;
        }
    }

    return zenkaku;
}

BOOL test_salt(int x) {
    BOOL b = 0; /* FALSE */;
    char salt[SALT_ARRAY_SIZE] = {' ', '!', '"', '#', '(', ')', '*', '+', ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~', '$', '%', '&', 0x27};
    int k;
    for(k = 0; k < SALT_ARRAY_SIZE; k++) {
        if(salt[k] == x) {
            b = TRUE;
        }

    }

    return b;
}

void set_byte_table(void) {
    byte_table['0']['0'] = 0x00;
    byte_table['0']['1'] = 0x01;
    byte_table['0']['2'] = 0x02;
    byte_table['0']['3'] = 0x03;
    byte_table['0']['4'] = 0x04;
    byte_table['0']['5'] = 0x05;
    byte_table['0']['6'] = 0x06;
    byte_table['0']['7'] = 0x07;
    byte_table['0']['8'] = 0x08;
    byte_table['0']['9'] = 0x09;
    byte_table['0']['A'] = 0x0A;
    byte_table['0']['B'] = 0x0B;
    byte_table['0']['C'] = 0x0C;
    byte_table['0']['D'] = 0x0D;
    byte_table['0']['E'] = 0x0E;
    byte_table['0']['F'] = 0x0F;
    byte_table['1']['0'] = 0x10;
    byte_table['1']['1'] = 0x11;
    byte_table['1']['2'] = 0x12;
    byte_table['1']['3'] = 0x13;
    byte_table['1']['4'] = 0x14;
    byte_table['1']['5'] = 0x15;
    byte_table['1']['6'] = 0x16;
    byte_table['1']['7'] = 0x17;
    byte_table['1']['8'] = 0x18;
    byte_table['1']['9'] = 0x19;
    byte_table['1']['A'] = 0x1A;
    byte_table['1']['B'] = 0x1B;
    byte_table['1']['C'] = 0x1C;
    byte_table['1']['D'] = 0x1D;
    byte_table['1']['E'] = 0x1E;
    byte_table['1']['F'] = 0x1F;
    byte_table['2']['0'] = 0x20;
    byte_table['2']['1'] = 0x21;
    byte_table['2']['2'] = 0x22;
    byte_table['2']['3'] = 0x23;
    byte_table['2']['4'] = 0x24;
    byte_table['2']['5'] = 0x25;
    byte_table['2']['6'] = 0x26;
    byte_table['2']['7'] = 0x27;
    byte_table['2']['8'] = 0x28;
    byte_table['2']['9'] = 0x29;
    byte_table['2']['A'] = 0x2A;
    byte_table['2']['B'] = 0x2B;
    byte_table['2']['C'] = 0x2C;
    byte_table['2']['D'] = 0x2D;
    byte_table['2']['E'] = 0x2E;
    byte_table['2']['F'] = 0x2F;
    byte_table['3']['0'] = 0x30;
    byte_table['3']['1'] = 0x31;
    byte_table['3']['2'] = 0x32;
    byte_table['3']['3'] = 0x33;
    byte_table['3']['4'] = 0x34;
    byte_table['3']['5'] = 0x35;
    byte_table['3']['6'] = 0x36;
    byte_table['3']['7'] = 0x37;
    byte_table['3']['8'] = 0x38;
    byte_table['3']['9'] = 0x39;
    byte_table['3']['A'] = 0x3A;
    byte_table['3']['B'] = 0x3B;
    byte_table['3']['C'] = 0x3C;
    byte_table['3']['D'] = 0x3D;
    byte_table['3']['E'] = 0x3E;
    byte_table['3']['F'] = 0x3F;
    byte_table['4']['0'] = 0x40;
    byte_table['4']['1'] = 0x41;
    byte_table['4']['2'] = 0x42;
    byte_table['4']['3'] = 0x43;
    byte_table['4']['4'] = 0x44;
    byte_table['4']['5'] = 0x45;
    byte_table['4']['6'] = 0x46;
    byte_table['4']['7'] = 0x47;
    byte_table['4']['8'] = 0x48;
    byte_table['4']['9'] = 0x49;
    byte_table['4']['A'] = 0x4A;
    byte_table['4']['B'] = 0x4B;
    byte_table['4']['C'] = 0x4C;
    byte_table['4']['D'] = 0x4D;
    byte_table['4']['E'] = 0x4E;
    byte_table['4']['F'] = 0x4F;
    byte_table['5']['0'] = 0x50;
    byte_table['5']['1'] = 0x51;
    byte_table['5']['2'] = 0x52;
    byte_table['5']['3'] = 0x53;
    byte_table['5']['4'] = 0x54;
    byte_table['5']['5'] = 0x55;
    byte_table['5']['6'] = 0x56;
    byte_table['5']['7'] = 0x57;
    byte_table['5']['8'] = 0x58;
    byte_table['5']['9'] = 0x59;
    byte_table['5']['A'] = 0x5A;
    byte_table['5']['B'] = 0x5B;
    byte_table['5']['C'] = 0x5C;
    byte_table['5']['D'] = 0x5D;
    byte_table['5']['E'] = 0x5E;
    byte_table['5']['F'] = 0x5F;
    byte_table['6']['0'] = 0x60;
    byte_table['6']['1'] = 0x61;
    byte_table['6']['2'] = 0x62;
    byte_table['6']['3'] = 0x63;
    byte_table['6']['4'] = 0x64;
    byte_table['6']['5'] = 0x65;
    byte_table['6']['6'] = 0x66;
    byte_table['6']['7'] = 0x67;
    byte_table['6']['8'] = 0x68;
    byte_table['6']['9'] = 0x69;
    byte_table['6']['A'] = 0x6A;
    byte_table['6']['B'] = 0x6B;
    byte_table['6']['C'] = 0x6C;
    byte_table['6']['D'] = 0x6D;
    byte_table['6']['E'] = 0x6E;
    byte_table['6']['F'] = 0x6F;
    byte_table['7']['0'] = 0x70;
    byte_table['7']['1'] = 0x71;
    byte_table['7']['2'] = 0x72;
    byte_table['7']['3'] = 0x73;
    byte_table['7']['4'] = 0x74;
    byte_table['7']['5'] = 0x75;
    byte_table['7']['6'] = 0x76;
    byte_table['7']['7'] = 0x77;
    byte_table['7']['8'] = 0x78;
    byte_table['7']['9'] = 0x79;
    byte_table['7']['A'] = 0x7A;
    byte_table['7']['B'] = 0x7B;
    byte_table['7']['C'] = 0x7C;
    byte_table['7']['D'] = 0x7D;
    byte_table['7']['E'] = 0x7E;
    byte_table['7']['F'] = 0x7F;
    byte_table['8']['0'] = 0x0080;
    byte_table['8']['1'] = 0x0081;
    byte_table['8']['2'] = 0x0082;
    byte_table['8']['3'] = 0x0083;
    byte_table['8']['4'] = 0x0084;
    byte_table['8']['5'] = 0x0085;
    byte_table['8']['6'] = 0x0086;
    byte_table['8']['7'] = 0x0087;
    byte_table['8']['8'] = 0x0088;
    byte_table['8']['9'] = 0x0089;
    byte_table['8']['A'] = 0x008A;
    byte_table['8']['B'] = 0x008B;
    byte_table['8']['C'] = 0x008C;
    byte_table['8']['D'] = 0x008D;
    byte_table['8']['E'] = 0x008E;
    byte_table['8']['F'] = 0x008F;
    byte_table['9']['0'] = 0x0090;
    byte_table['9']['1'] = 0x0091;
    byte_table['9']['2'] = 0x0092;
    byte_table['9']['3'] = 0x0093;
    byte_table['9']['4'] = 0x0094;
    byte_table['9']['5'] = 0x0095;
    byte_table['9']['6'] = 0x0096;
    byte_table['9']['7'] = 0x0097;
    byte_table['9']['8'] = 0x0098;
    byte_table['9']['9'] = 0x0099;
    byte_table['9']['A'] = 0x009A;
    byte_table['9']['B'] = 0x009B;
    byte_table['9']['C'] = 0x009C;
    byte_table['9']['D'] = 0x009D;
    byte_table['9']['E'] = 0x009E;
    byte_table['9']['F'] = 0x009F;
    byte_table['A']['0'] = 0x00A0;
    byte_table['A']['1'] = 0x00A1;
    byte_table['A']['2'] = 0x00A2;
    byte_table['A']['3'] = 0x00A3;
    byte_table['A']['4'] = 0x00A4;
    byte_table['A']['5'] = 0x00A5;
    byte_table['A']['6'] = 0x00A6;
    byte_table['A']['7'] = 0x00A7;
    byte_table['A']['8'] = 0x00A8;
    byte_table['A']['9'] = 0x00A9;
    byte_table['A']['A'] = 0x00AA;
    byte_table['A']['B'] = 0x00AB;
    byte_table['A']['C'] = 0x00AC;
    byte_table['A']['D'] = 0x00AD;
    byte_table['A']['E'] = 0x00AE;
    byte_table['A']['F'] = 0x00AF;
    byte_table['B']['0'] = 0x00B0;
    byte_table['B']['1'] = 0x00B1;
    byte_table['B']['2'] = 0x00B2;
    byte_table['B']['3'] = 0x00B3;
    byte_table['B']['4'] = 0x00B4;
    byte_table['B']['5'] = 0x00B5;
    byte_table['B']['6'] = 0x00B6;
    byte_table['B']['7'] = 0x00B7;
    byte_table['B']['8'] = 0x00B8;
    byte_table['B']['9'] = 0x00B9;
    byte_table['B']['A'] = 0x00BA;
    byte_table['B']['B'] = 0x00BB;
    byte_table['B']['C'] = 0x00BC;
    byte_table['B']['D'] = 0x00BD;
    byte_table['B']['E'] = 0x00BE;
    byte_table['B']['F'] = 0x00BF;
    byte_table['C']['0'] = 0x00C0;
    byte_table['C']['1'] = 0x00C1;
    byte_table['C']['2'] = 0x00C2;
    byte_table['C']['3'] = 0x00C3;
    byte_table['C']['4'] = 0x00C4;
    byte_table['C']['5'] = 0x00C5;
    byte_table['C']['6'] = 0x00C6;
    byte_table['C']['7'] = 0x00C7;
    byte_table['C']['8'] = 0x00C8;
    byte_table['C']['9'] = 0x00C9;
    byte_table['C']['A'] = 0x00CA;
    byte_table['C']['B'] = 0x00CB;
    byte_table['C']['C'] = 0x00CC;
    byte_table['C']['D'] = 0x00CD;
    byte_table['C']['E'] = 0x00CE;
    byte_table['C']['F'] = 0x00CF;
    byte_table['D']['0'] = 0x00D0;
    byte_table['D']['1'] = 0x00D1;
    byte_table['D']['2'] = 0x00D2;
    byte_table['D']['3'] = 0x00D3;
    byte_table['D']['4'] = 0x00D4;
    byte_table['D']['5'] = 0x00D5;
    byte_table['D']['6'] = 0x00D6;
    byte_table['D']['7'] = 0x00D7;
    byte_table['D']['8'] = 0x00D8;
    byte_table['D']['9'] = 0x00D9;
    byte_table['D']['A'] = 0x00DA;
    byte_table['D']['B'] = 0x00DB;
    byte_table['D']['C'] = 0x00DC;
    byte_table['D']['D'] = 0x00DD;
    byte_table['D']['E'] = 0x00DE;
    byte_table['D']['F'] = 0x00DF;
    byte_table['E']['0'] = 0x00E0;
    byte_table['E']['1'] = 0x00E1;
    byte_table['E']['2'] = 0x00E2;
    byte_table['E']['3'] = 0x00E3;
    byte_table['E']['4'] = 0x00E4;
    byte_table['E']['5'] = 0x00E5;
    byte_table['E']['6'] = 0x00E6;
    byte_table['E']['7'] = 0x00E7;
    byte_table['E']['8'] = 0x00E8;
    byte_table['E']['9'] = 0x00E9;
    byte_table['E']['A'] = 0x00EA;
    byte_table['E']['B'] = 0x00EB;
    byte_table['E']['C'] = 0x00EC;
    byte_table['E']['D'] = 0x00ED;
    byte_table['E']['E'] = 0x00EE;
    byte_table['E']['F'] = 0x00EF;
    byte_table['F']['0'] = 0x00F0;
    byte_table['F']['1'] = 0x00F1;
    byte_table['F']['2'] = 0x00F2;
    byte_table['F']['3'] = 0x00F3;
    byte_table['F']['4'] = 0x00F4;
    byte_table['F']['5'] = 0x00F5;
    byte_table['F']['6'] = 0x00F6;
    byte_table['F']['7'] = 0x00F7;
    byte_table['F']['8'] = 0x00F8;
    byte_table['F']['9'] = 0x00F9;
    byte_table['F']['A'] = 0x00FA;
    byte_table['F']['B'] = 0x00FB;
    byte_table['F']['C'] = 0x00FC;
    byte_table['F']['D'] = 0x00FD;
    byte_table['F']['E'] = 0x00FE;
    byte_table['F']['F'] = 0x00FF;
}

int main(int argc, char *argv[]) {
    BOOL check[ CHARACTER_COUNT + 1 ];
    char *p;
    char buffer[BUFFER_SIZE];
    char c[BUFFER_SIZE];
    char byte_code;
    const char abandon[] = " | \r\n";
    const char new_line[] = "\r\n";
    FILE *binary;
    FILE *text;
    int q, x, y;

    switch(argc) {
    case 2:
        if(! strcmp(argv[1], "-V") ) {
            fprintf(stderr, "sjis Version %1.2f\n\n", VERSION);
            fprintf(stderr, "sjis \n");
            fprintf(stderr, "sjis file_name\n");
            fprintf(stderr, "sjis -V\n");
            return 1;
        }

        text = fopen(argv[1], "rt");
        if(text == NULL) {
            perror(argv[1]);
            return 1;
        }

        break;
    default:
        text = fopen(VOYA_LOG, "rt");
        if(text == NULL) {
            perror(VOYA_LOG);
            return 1;
        }

        break;
    }

    set_byte_table();
    binary = fopen(BINARY_FILE, "wb");

    for(;;) {
        p = fgets(c, sizeof(c), text);
        if(p == NULL) {
            break;
        }

        p = strstr(c, FROM_TO);
        if(p == NULL) {
            continue;
        }

        chopped(c);
/*
voya.log includes "raw key -> tori" lines.
D3CDC4DBA9D8A49C -> Amin//dS0M
E4BB98E7905F96F1 -> Ruda/tI7a2
*/
        p = &c[0];
        for(q = 0; q < CHARACTER_COUNT; q++) {
            x = (int)p[0];
            y = (int)p[1];
            byte_code = (char)byte_table[x][y];
            buffer[q] = byte_code;
            check[q] = test_salt(byte_code);

            p += 2;
        }

        buffer[CHARACTER_COUNT] = '\0';
        fwrite(buffer, CHARACTER_COUNT, 1, binary);

        if( zenkaku_fragment(buffer) ) {
            fwrite(abandon, strlen(abandon), 1, binary);
            continue;
        } else {
            if(*p == ' ') {
/* If raw-key length is sixteen, voya uses ".." as salt; however 
the server may use the second and third bytes (buffer[1] and buffer[2]) as salt. 
*/
                if(check[1] || check[2]) {
                    fwrite(abandon, strlen(abandon), 1, binary);
                    continue;
                }
            } else {
/*  If raw-key length is eighteen, the second and third bytes (buffer[1] and buffer[2]) are used as salt.
bc4e595b2a76407bNY -> Shizuoka8I 
N (4Eh)
Y (59h)
*/
                p += 2;
            }

        }

        fwrite(p, strlen(p), 1, binary);
        fwrite(new_line, strlen(new_line), 1, binary);
    }

    fclose(binary);
    fclose(text);
    return 0;
}
