#include "queue.h"

char *queue[QUEUE_ARRAY_SIZE]; /* queue_size() <= QUEUE_ARRAY_SIZE - 1 */
int queue_begin, queue_end;

int display(void) {
    int item, k;
    int size = queue_size();

    item = queue_begin;
    for(k = 0; k < size; k++) {
        puts(queue[item]);
        item = next(item);
    }

    return 0;
}

int empty(void) {
    return queue_begin == queue_end;
}

int next(int index) {
    return (index + 1) % QUEUE_ARRAY_SIZE;
}

int pop_front(void) {
    if( empty() ) {
        return EMPTY_ERROR;
    }

    free(queue[queue_begin]);
    queue[queue_begin] = NULL;
    queue_begin = next(queue_begin);
    return QUEUE_SHRUNK;
}

int produce_ring_buffer(int size) {
    char BLANK[2] = " ";
    int k;

    clear();
    for(k = 1; k <= size; k++) {
        push_back(BLANK);
    }

    return queue_size();
}

int push_back(char *p) {
    if( next(queue_end) == queue_begin ) {
        return FULL_HOUSE_ERROR;
    }

    queue[queue_end] = (char *)malloc(ITEM_SIZE);
    if(queue[queue_end] == NULL) {
        return MALLOC_ERROR;
    }

    strcpy(queue[queue_end], p);
    queue_end = next(queue_end);
    return QUEUE_GROWN;
}

int queue_size(void) {
    int size;

    if(queue_begin < queue_end) {
        size = queue_end - queue_begin;
    } else if(queue_begin == queue_end) {
        size = 0;
    } else {
        size = queue_end + QUEUE_ARRAY_SIZE - queue_begin;
    }

    return size;
}

void clear(void) {
    int k;
    int repetition = queue_size();
    for(k = 0; k < repetition; k++) {
        pop_front();
    }
}

