#include "voyaSetting.h"

extern VOYA_ERROR *ve;
VOYA_SETTING voyaSetting;
VOYA_SETTING *vs = &voyaSetting;

static double EPSILON = 0.5;

BOOL copyLogFileName(char *sequence, int end) {
    char *edited_line;
    int i = end;
    while(i > 0) {
        if(sequence[i] == ',') {
            break;
        }

        --i;
    }

    if(sequence[i] != ',') {
        voyaMessageFlush("Comma should come before log file.");
        return 0; /* FALSE */
    }

    for(;;) {
        if(sequence[i] == ' ') {
            ++i;
            continue;
        } else if(sequence[i] == ',') {
            ++i;
            continue;
        } else {
            break;
        }
    }

    edited_line = & vs->edited[LOG_LINE][0];
    strcpy(edited_line, &sequence[i]);
    strcpy(vs->log_file_name, &sequence[i]);
    return TRUE;
}

BOOL copyRequest(char *sequence, int end) {
    char *edited_line;
    int i = end;
    while(i > 0) {
        if(sequence[i] == ',') {
            break;
        }

        --i;
    }

    if(sequence[i] != ',') {
        voyaMessageFlush("Comma should come before request.");
        return 0; /* FALSE */
    }

    for(;;) {
        if(sequence[i] == ',') {
            ++i;
            continue;
        } else if(sequence[i] == ' ') {
            ++i;
            continue;
        } else {
            break;
        }
    }

    edited_line = & vs->edited[REQUEST_LINE][0];
    strcpy(edited_line, &sequence[i]);
    strcpy(vs->request, &sequence[i]);
    return TRUE;
}

BOOL copySalt(char *sequence, int end) {
    char *edited_line;
    int i = end;
    while(i > 0) {
        if(sequence[i] == ',') {
            break;
        }

        --i;
    }

    if(sequence[i] != ',') {
        voyaMessageFlush("Comma should come before salt.");
        return 0; /* FALSE */
    }

    for(;;) {
        if(sequence[i] == ',') {
            ++i;
            continue;
        } else if(sequence[i] == ' ') {
            ++i;
            continue;
        } else {
            break;
        }
    }

    edited_line = & vs->edited[SALT_LINE][0];
    strcpy(edited_line, &sequence[i]);
    strcpy(vs->salt, &sequence[i]);
    return TRUE;
}

BOOL readLabels(void) {
    char * cells[CSV_COLUMN_COUNT][ LABEL_COUNT + 1 ];
    char *base;
    char *p;
    char c[SMALL_BUFFER_SIZE];
    const char HEADER_DATA[] = "Data";
    const char HEADER_WORD[] = "Word";
    const int COMMA = ',';
    const int LABEL_COUNT_PLUS = LABEL_COUNT + 1;
    FILE *file;
    int k, m, line;

    file = fopen(VOYA_CSV, "rt");
    if(file == NULL) {
        voyaMessageFlush("readLabels(), file = fopen(VOYA_CSV, \"rt\") ");
        return 0; /* FALSE */
    }

    for(k = 0; k < CSV_COLUMN_COUNT; k++) {
        for(m = 0; m < LABEL_COUNT_PLUS; m++) {
            cells[k][m] = malloc(SMALL_BUFFER_SIZE);
            if(cells[k][m] == NULL) {
                voyaMessageFlush("cells[k][m] == NULL");
                return 0; /* FALSE */
            }
        }
    }

// Reading and copying to produce a matrix
    strcpy(cells[0][0], HEADER_WORD);
    strcpy(cells[1][0], HEADER_DATA);
    for(line = 1; line <= LABEL_COUNT; line += 1) {
        p = fgets(c, SMALL_BUFFER_SIZE, file);
        if(p == NULL) {
            break;
        }

        for(k = 0; k < CSV_COLUMN_COUNT; k++) {
            base = p;
            p = strchr(base, COMMA);
            if(p != NULL) {
                *p = '\0';
                p++;
                strcpy(cells[k][line], base);    
                continue;
            } else {
                strcpy(cells[k][line], base);
                break;
            }

        }

    }

    fclose(file);

// Storing labels
    p = & vs->label[0][0];
    strcpy(p, HEADER_WORD);
    for(line = 1; line <= LABEL_COUNT; line += 1) {
        p = & vs->label[line][0];
        strcpy(p, cells[0][line]);
    }

    for(k = 0; k < CSV_COLUMN_COUNT; k++) {
        for(m = 0; m < LABEL_COUNT_PLUS; m++) {
            free(cells[k][m]);
        }
    }

    return TRUE;
}

BOOL removeSalt(void) {
    char *p;
    char *q;
    char buffer[BUFFER_SIZE];
    char c[BUFFER_SIZE];
    char COMMA[] = {',', '\0'};
    char NEW_LINE[] = {'\n', '\0'};
    const int COMMA_SIZE = 1;
    FILE *file;
    int line = 0;

    file = fopen(VOYA_CSV, "rt");
    if(file == NULL) {
        return 0; /* FALSE */
    }

    buffer[0] = '\0';
    for(;;) {
        p = fgets(c, BUFFER_SIZE, file);
        if(p == NULL) {
            break;
        }

        line += 1;

        if(line == SALT_LINE) {
/* Removing salt from VOYA_CSV */
            q = strstr(c, COMMA);
            if(q != NULL) {
                q += COMMA_SIZE;
                *q = '\0';
                strcat(buffer, c);
                strcat(buffer, NEW_LINE);
                continue;
            }

        }

        strcat(buffer, c);
    }

    fclose(file);

/* Writing */
    file = fopen(VOYA_CSV, "wt");
    fputs(buffer, file);
    fclose(file);
    return TRUE;
}

BOOL settingBy(const char *fileName) {
    BOOL used;
    char *b;
    char *edited_line;
    char buffer[SMALL_BUFFER_SIZE];
    char sequence[BUFFER_SIZE];
    FILE *voyaCSV;
    int a[ARRAY_LENGTH];
    int aeim[AEIM_CHARACTER] = {'a', 'e', 'i', 'm', 'q', 'u', 'y', 'C', 'G', 'K', 'O', 'S', 'W'};
    int CGKO[CGKO_CHARACTER] = {'C', 'G', 'K', 'O', 'S', 'W'};
    int k, length, m, tail, u;
    int line = LEFTMOST_LINE;

    voyaCSV = fopen(fileName, "rt");
    if(voyaCSV == NULL) {
        voyaMessageFlush("voyaCSV == NULL");
        return 0; /* FALSE */
    }

// request
    b = fgets(sequence, BUFFER_SIZE, voyaCSV);
    if(b == NULL) {
        fclose(voyaCSV);
        voyaMessageFlush("request, b = fgets(sequence, BUFFER_SIZE, voyaCSV)");
        return 0; /* FALSE */
    }

    length = removeComment(sequence);
    if( copyRequest(sequence, length) ) {
        if( metaCharacterIncluded(vs->request) ) {
            vs->regular_expression = TRUE;

/* 
012345678901
^[Rr]uda[/.]
*/
            u = strlen(vs->request);
            if(vs->request[0] == '^') {
                if(vs->request[1] == '[') {
                    if(vs->request[4] == ']') {
                        if(vs->request[ u - 1 ] == ']') {
                            if(vs->request[ u - 4 ] == '[') {
                                vs->dragon_seek = TRUE;
                            } else {
                                vs->dragon_seek = 0; /* FALSE */
                            }
                        } else {
                            vs->dragon_seek = 0; /* FALSE */
                        }
                    } else {
                        vs->dragon_seek = 0; /* FALSE */
                    }
                } else {
                    vs->dragon_seek = 0; /* FALSE */
                }
            } else {
                vs->dragon_seek = 0; /* FALSE */
            }

        } else {
            vs->regular_expression = 0; /* FALSE */
            vs->dragon_seek = 0; /* FALSE */
        }
    } else {
        fclose(voyaCSV);
        voyaMessageFlush("copyRequest(sequence, length)");
        return 0; /* FALSE */
    }

// log file name
    b = fgets(sequence, BUFFER_SIZE, voyaCSV);
    if(b == NULL) {
        fclose(voyaCSV);
        voyaMessageFlush("log file name, b = fgets(sequence, BUFFER_SIZE, voyaCSV)");
        return 0; /* FALSE */
    }

    length = removeComment(sequence);
    if(! copyLogFileName(sequence, length) ) {
        fclose(voyaCSV);
        voyaMessageFlush("copyLogFileName(sequence, length)");
        return 0; /* FALSE */
    }

// Converting strings to integers
    for(k = 0; k < ARRAY_LENGTH; k++) {
        b = fgets(sequence, BUFFER_SIZE, voyaCSV);
        if(b == NULL) {
            break;
        }

        length = removeComment(sequence);
        a[k] = commaShouldComeBeforeInteger(sequence, length, k); // vs->buffer
        edited_line = & vs->edited[line][0];
        strcpy(edited_line, vs->buffer);
        line += 1;
    }

// salt
    b = fgets(sequence, BUFFER_SIZE, voyaCSV);
    if(b == NULL) {
        fclose(voyaCSV);
        voyaMessageFlush("salt, voyaCSV.getline(sequence, BUFFER_SIZE)");
        return 0; /* FALSE */
    }

    length = removeComment(sequence);
    if( copySalt(sequence, length) ) {
        if( testSalt(vs->salt[0]) ) {
            if( testSalt(vs->salt[1]) ) {
                if(vs->leftmost_salt) {
                    ;
                } else {
                    voyaMessageFlush("testSalt(vs->salt[0]) && testSalt(vs->salt[1]) is inconsistent with vs->leftmost_salt.");
                }

            }

        }

    } else {
        fclose(voyaCSV);
        voyaMessageFlush("copySalt(sequence, length)");
        return 0; /* FALSE */
    }

// weight
    vs->ascii_weight = a[4];
    vs->capital_omega_weight = a[5];
    vs->capital_ya_weight = a[6];
    vs->capital_z_weight = a[7];
    vs->cap_weight = a[8];
    vs->circle_weight = a[9];
    vs->cup_weight = a[10];
    vs->double_integral_weight = a[11];
    vs->exists_weight = a[12];
    vs->hankaku_katakana_weight = a[13];
    vs->heisei_weight = a[14];
    vs->hiragana_weight = a[15];
    vs->jis_level_one_kanji_weight = a[16];
    vs->jis_level_two_kanji_weight = a[17];
    vs->katakana_weight = a[18];
    vs->keisen_weight = a[19];
    vs->nine_weight = a[20];
    vs->omega_weight = a[21];
    vs->paragraph_weight = a[22];
    vs->parallel_weight = a[23];
    vs->roman_weight = a[24];
    vs->square_meter_weight = a[25];
    vs->ya_weight = a[26];
    vs->z_weight = a[27];

// constants
    vs->interim_report = a[2];
    vs->idling = a[3];

    if(vs->regular_expression) {
        if(vs->dragon_seek) {
            vs->case_sensitive = TRUE;
            vs->leftmost = 1;
            vs->lower_case = 0; /* FALSE */
        } else {
            vs->case_sensitive = 0; /* FALSE */
            vs->leftmost = 0;
            vs->lower_case = 0; /* FALSE */
        }

    } else {
        vs->case_sensitive = (BOOL)a[1];
        vs->leftmost = a[0];
        if(vs->leftmost == -1) { // rightmost
            u = strlen(vs->request);
            u -= 1;
            tail = (int)vs->request[u]; // the last character of request
            used = tenthCharacterOfTori(tail);
            if(used) {
                useLowerCaseUnlessCaseSensitive();
            } else {
                voyaMessageFlush(SHOULD_BE_USED_AS_TENTH_CHARACTER);
                if(vs->case_sensitive) {
                    fclose(voyaCSV);
                    return 0; /* FALSE */
                } else {
// Correcting automatically
                    used = 0; /* FALSE */
                    for(m = 0; m < AEIM_CHARACTER; ++m) {
                        if(tail == aeim[m]) {
                            used = TRUE;
                            break;
                        }
                    }

                    if(used) {
                        used = 0; /* FALSE */
                        for(m = 0; m < CGKO_CHARACTER; ++m) {
                            if(tail == CGKO[m]) {
                                used = TRUE;
                                break;
                            }
                        }

                        if(used) { // C', 'G', 'K', 'O', 'S', 'W'
                            vs->lower_case = TRUE;
                            k = 0;
                            for(b = & vs->request[0]; *b != '\0'; ++b) {
                                buffer[k++] = tolower(*b);
                            }

                            buffer[k] = '\0';
                            strcpy(vs->request, buffer);
                        } else { // 'a', 'e', 'i', 'm', 'q', 'u', 'y' 
                            vs->lower_case = 0; /* FALSE */
                            k = 0;
                            for(b = & vs->request[0]; *b != '\0'; ++b) {
                                buffer[k++] = toupper(*b);
                            }

                            buffer[k] = '\0';
                            strcpy(vs->request, buffer);
                        }
                    } else {
                        fclose(voyaCSV);
                        voyaMessageFlush("int aeim[AEIM_CHARACTER] = {'a', 'e', 'i', 'm', 'q', 'u', 'y', 'C', 'G', 'K', 'O', 'S', 'W'}; if(used) ");
                        return 0; /* FALSE */
                    }
                }
            }
        } else {
            useLowerCaseUnlessCaseSensitive();
        }
    }

    fclose(voyaCSV);
    return TRUE;
}

BOOL testSalt(int x) {
    BOOL b = 0; /* FALSE */
    char salt[SALT_ARRAY_SIZE] = {' ', '!', '"', '#', '(', ')', '*', '+', ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~', '$', '%', '&', 0x27};
    int k;
    for(k = 0; k < SALT_ARRAY_SIZE; k++) {
        if(salt[k] == x) {
            b = TRUE;
        }

    }

    return b;
}

BOOL writeLabelsEditedLines(void) {
    char * cells[CSV_COLUMN_COUNT][ LABEL_COUNT + 1 ];
    char *edited_line;
    char *p;
    const char EMPTY[1] = {'\0'};
    const int COMMA = ',';
    const int LABEL_COUNT_PLUS = LABEL_COUNT + 1;
    const int NEW_LINE = '\n';
    FILE *file;
    int delimiter, k, m, line;

// malloc()
    for(k = 0; k < CSV_COLUMN_COUNT; k++) {
        for(m = 0; m < LABEL_COUNT_PLUS; m++) {
            cells[k][m] = malloc(SMALL_BUFFER_SIZE);
            if(cells[k][m] == NULL) {
                voyaMessageFlush("cells[k][m] == NULL");
                return 0; /* FALSE */
            }
        }
    }

    strcpy(cells[0][0], EMPTY);
    strcpy(cells[1][0], EMPTY);

// Copying to produce a matrix
    for(line = 1; line <= LABEL_COUNT; line += 1) {
        edited_line = & vs->edited[line][0];
        strcpy(cells[1][line], edited_line);

        p = & vs->label[line][0];
        strcpy(cells[0][line], p);
    }

// Writing
    file = fopen(VOYA_CSV, "wt");
    for(line = 1; line <= LABEL_COUNT; line += 1) {
        for(k = 0; k < CSV_COLUMN_COUNT; k++) {
            switch(k) {
            case 0:
                delimiter = COMMA;
                break;
            default:
                delimiter = NEW_LINE;
                break;
            }

            fprintf(file, "%s%c", cells[k][line], delimiter);
        }
    }

    fclose(file);

// free()
    for(k = 0; k < CSV_COLUMN_COUNT; k++) {
        for(m = 0; m < LABEL_COUNT_PLUS; m++) {
            free(cells[k][m]);
        }
    }

    return TRUE;
}

int commaShouldComeBeforeInteger(char *sequence, int length, int index) {
    int i = length;
    while(i > 0) {
        if(sequence[i] == ',') {
            break;
        }

        --i;
    }

    if(sequence[i] != ',') {
        switch(index) {
        case 0:
            voyaMessageFlush("Comma should come before the leftmost value.");
            break;
        case 1:
            voyaMessageFlush("Comma should come before the case_sensitive value.");
            break;
        case 2:
            voyaMessageFlush("Comma should come before the interim report value.");
            break;
        case 3:
            voyaMessageFlush("Comma should come before the idling value.");
            break;
        default:
            voyaMessage("Comma should come before weight: ");
            voyaCountFlush(i);
            break;
        }

        return 0; /* FALSE */
    }

    for(;;) {
        if(sequence[i] == ',') {
            ++i;
            continue;
        } else if(sequence[i] == ' ') {
            ++i;
            continue;
        } else {
            break;
        }
    }

    strcpy(vs->buffer, &sequence[i]);
    i = cstring_to_integer(vs->buffer);
    return i;
}

void editedLinePrint(int line, int a) {
    char *edited_line = & vs->edited[line][0];
    sprintf(edited_line, "%d", a);
}

void normalizing(double sum, double aim) {
    double d, rate;
    double weight[WEIGHT_ARRAY_LENGTH];
    int i, n;
    int a[WEIGHT_ARRAY_LENGTH];

    a[0] = vs->ascii_weight;
    a[1] = vs->capital_omega_weight;
    a[2] = vs->capital_ya_weight;
    a[3] = vs->capital_z_weight;
    a[4] = vs->cap_weight;
    a[5] = vs->circle_weight;
    a[6] = vs->cup_weight;
    a[7] = vs->double_integral_weight;
    a[8] = vs->exists_weight;
    a[9] = vs->hankaku_katakana_weight;
    a[10] = vs->heisei_weight;
    a[11] = vs->hiragana_weight;
    a[12] = vs->jis_level_one_kanji_weight;
    a[13] = vs->jis_level_two_kanji_weight;
    a[14] = vs->katakana_weight;
    a[15] = vs->keisen_weight;
    a[16] = vs->nine_weight;
    a[17] = vs->omega_weight;
    a[18] = vs->paragraph_weight;
    a[19] = vs->parallel_weight;
    a[20] = vs->roman_weight;
    a[21] = vs->square_meter_weight;
    a[22] = vs->ya_weight;
    a[23] = vs->z_weight;

    for(n = 0; n < REPEATEDLY_NORMALIZING; n++) {
        for(i = 0; i < WEIGHT_ARRAY_LENGTH; i++) {
            weight[i] = (double) a[i];
        }

        rate = aim/sum;
        for(i = 0; i < WEIGHT_ARRAY_LENGTH; ++i) {
            d = rate * weight[i];
            if(1 < d) {
                a[i] = (int)d;
            } else if(EPSILON < d) {
                a[i] = 1;
            } else {
                a[i] = 0;
            }

        }

        vs->ascii_weight = a[0];
        vs->capital_omega_weight = a[1];
        vs->capital_ya_weight = a[2];
        vs->capital_z_weight = a[3];
        vs->cap_weight = a[4];
        vs->circle_weight = a[5];
        vs->cup_weight = a[6];
        vs->double_integral_weight = a[7];
        vs->exists_weight = a[8];
        vs->hankaku_katakana_weight = a[9];
        vs->heisei_weight = a[10];
        vs->hiragana_weight = a[11];
        vs->jis_level_one_kanji_weight = a[12];
        vs->jis_level_two_kanji_weight = a[13];
        vs->katakana_weight = a[14];
        vs->keisen_weight = a[15];
        vs->nine_weight = a[16];
        vs->omega_weight = a[17];
        vs->paragraph_weight = a[18];
        vs->parallel_weight = a[19];
        vs->roman_weight = a[20];
        vs->square_meter_weight = a[21];
        vs->ya_weight = a[22];
        vs->z_weight = a[23];
        sum = shiftJisKanaAscii();
    }

    editedLinePrint(7, a[0]);
    editedLinePrint(8, a[1]);
    editedLinePrint(9, a[2]);
    editedLinePrint(10, a[3]);
    editedLinePrint(11, a[4]);
    editedLinePrint(12, a[5]);
    editedLinePrint(13, a[6]);
    editedLinePrint(14, a[7]);
    editedLinePrint(15, a[8]);
    editedLinePrint(16, a[9]);
    editedLinePrint(17, a[10]);
    editedLinePrint(18, a[11]);
    editedLinePrint(19, a[12]);
    editedLinePrint(20, a[13]);
    editedLinePrint(21, a[14]);
    editedLinePrint(22, a[15]);
    editedLinePrint(23, a[16]);
    editedLinePrint(24, a[17]);
    editedLinePrint(25, a[18]);
    editedLinePrint(26, a[19]);
    editedLinePrint(27, a[20]);
    editedLinePrint(28, a[21]);
    editedLinePrint(29, a[22]);
    editedLinePrint(30, a[23]);
}

void recovery(void) {
    FILE *file = fopen(VOYA_CSV, "wt");
    fputs("Request,ase\n", file);
    fputs("Log,voya.log\n", file);
    fputs("Leftmost,0\n", file);
    fputs("Case-sensitive,0\n", file);
    fputs("Report,524287\n", file);
    fputs("Idling,400\n", file);
    fputs("ASCII @,143\n", file);
    fputs("Greek ,1\n", file);
    fputs("Cyrillic A,1\n", file);
    fputs("f,1\n", file);
    fputs(",1\n", file);
    fputs(",1\n", file);
    fputs(",1\n", file);
    fputs(",1\n", file);
    fputs(",1\n", file);
    fputs(",143\n", file);
    fputs("~,1\n", file);
    fputs(",1\n", file);
    fputs(",1\n", file);
    fputs(",1\n", file);
    fputs("A,1\n", file);
    fputs(",1\n", file);
    fputs("T,1\n", file);
    fputs("Greek ,1\n", file);
    fputs(",1\n", file);
    fputs(",1\n", file);
    fputs("Roman Z,1\n", file);
    fputs("Unit l,1\n", file);
    fputs("Cyrillic y,1\n", file);
    fputs(",1\n", file);
    fputs("salt,\n", file);
    fputs(",\n", file);
    fputs(",\n", file);
    fputs(",\n", file);
    fclose(file);
}

void useLowerCaseUnlessCaseSensitive(void) {
    char *b;
    char buffer[SMALL_BUFFER_SIZE];
    int k = 0;

    if(vs->case_sensitive) {
        vs->lower_case = 0; /* FALSE */
    } else {
        vs->lower_case = TRUE;
        for(b = & vs->request[0]; *b != '\0'; ++b) {
            buffer[k++] = tolower(*b);
        }

        buffer[k] = '\0';
        strcpy(vs->request, buffer);
    }
}

void useNoCharacterButASCII(void) {
    char data1[2] = "0";
    char data2[2] = "0";
    char data3[2] = "0";
    char data4[2] = "0";
    char data5[2] = "0";
    char data6[2] = "0";
    char data7[2] = "0";
    char data8[2] = "0";
    char data9[2] = "0";
    char data10[2] = "0";
    char data11[4] = "315"; // ASCII
    char data12[2] = "0";
    char data13[2] = "0";
    char data14[2] = "0";
    char data15[2] = "0";
    char data16[2] = "0";
    char data17[2] = "0";
    char data18[2] = "0";
    char data19[2] = "0";
    char data20[2] = "0";
    char data21[2] = "0";
    char data22[2] = "0";
    char data23[2] = "0";
    char data24[2] = "0";

    char *edited_line;
    edited_line = & vs->edited[19][0]; strcpy(edited_line, data1); // kanji-1
    edited_line = & vs->edited[20][0]; strcpy(edited_line, data2); // kanji-2
    edited_line = & vs->edited[13][0]; strcpy(edited_line, data3);
    edited_line = & vs->edited[17][0]; strcpy(edited_line, data4);
    edited_line = & vs->edited[22][0]; strcpy(edited_line, data5); // keisen
    edited_line = & vs->edited[26][0]; strcpy(edited_line, data6);
    edited_line = & vs->edited[16][0]; strcpy(edited_line, data7); // hankaku
    edited_line = & vs->edited[28][0]; strcpy(edited_line, data8);
    edited_line = & vs->edited[21][0]; strcpy(edited_line, data9);
    edited_line = & vs->edited[18][0]; strcpy(edited_line, data10); // hiragana
    edited_line = & vs->edited[7][0]; strcpy(edited_line, data11); // ASCII
    edited_line = & vs->edited[10][0]; strcpy(edited_line, data12);
    edited_line = & vs->edited[9][0]; strcpy(edited_line, data13);
    edited_line = & vs->edited[8][0]; strcpy(edited_line, data14);
    edited_line = & vs->edited[30][0]; strcpy(edited_line, data15);
    edited_line = & vs->edited[29][0]; strcpy(edited_line, data16);
    edited_line = & vs->edited[24][0]; strcpy(edited_line, data17);
    edited_line = & vs->edited[27][0]; strcpy(edited_line, data18);
    edited_line = & vs->edited[23][0]; strcpy(edited_line, data19);
    edited_line = & vs->edited[12][0]; strcpy(edited_line, data20);
    edited_line = & vs->edited[25][0]; strcpy(edited_line, data21);
    edited_line = & vs->edited[14][0]; strcpy(edited_line, data22); // roman
    edited_line = & vs->edited[15][0]; strcpy(edited_line, data23);
    edited_line = & vs->edited[11][0]; strcpy(edited_line, data24);

    writeLabelsEditedLines();
}

